# pyright: reportImportCycles=false

from __future__ import annotations
from enum import Enum
from pathlib import Path
from science_signal import Signal  # pyright: ignore[reportMissingTypeStubs]

from backscattering_analyzer.benchfactory import BenchFactory

from .projection import Projection
from .sensitivity import Sensitivity
from .metadatas.acquisition import Acquisition as mAcquisition
from .metadatas.bench import Bench as mBench
from .bench import Bench


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .measure import Measure
    from .channel import Channel
    from matplotlib.axes import Axes


class AcquisitionType(Enum):
    """
    Type of the acquisition
    """

    REFERENCE = "reference"
    EXCITED = "excited"


class Acquisition:
    """
    A data acquisition as a set of paths and files

    Attributes
    ----------
    folder: Path
        Path to the folder containing all data related to this acquisition
    amplitude: float
        Amplitude's signal injected for this acquisition in unknown unit (useless for now)
    duration: int
        Duration of this acquisition in second
    frequency: float
        Frequency's signal injected for this acquisition in Hz (useless for now)
    measure: Measure
        Measure containing this acquisition
    metadata: metadatas.acquisition.Acquisition | None
        Metadata associated to this acquisition
    name: str
        Name of this acquisition
    projection: Projection
        Projection associated to this acquisition
    sensitivities: list[Sensitivity] | None
        Sensitivities associated to this acquisition
    time: int
        Start time of this acquisition in GPS time
    type: AcquisitionType
        Type of this acquisition
    """

    def __init__(
        self,
        measure: Measure,
        name: str,
        model_file: Path,
    ) -> None:
        """
        Initializer of Acquisition

        Parameters
        ----------
        measure: Measure
            Measure containing this acquisition
        model_file: Path
            File containing the model used to compute projection
        name: str
            Name of this acquisition
        """
        folder = measure.folder / name

        if not folder.is_dir():
            raise Exception(_("{} is not a folder").format(folder))

        self.folder: Path = folder
        self.measure: Measure = measure
        self.name: str = name

        self.metadata: mAcquisition | None = None
        self.benches: list[Bench] = []
        self.amplitude: float | None = None
        self.duration: int | None = None
        self.frequency: float | None = None
        self.projection: Projection | None = None
        self.sensitivities: list[Sensitivity] | None = None
        self.time: int | None = None
        self.type: AcquisitionType | None = None

        self.load(model_file)

    def load(self, model_file: Path) -> None:
        """
        Load attributes from metadata

        Parameters
        ----------
        model_file: Path
            File containing the model used to compute projection
        """
        if self.measure.metadata is None:
            raise Exception(_("metadata not set"))
        for acquisition in self.measure.metadata.acquisitions:
            if acquisition.name == self.name:
                self.metadata = acquisition

                self.amplitude = acquisition.amplitude
                self.duration = acquisition.duration
                self.frequency = acquisition.frequency
                self.time = acquisition.time
                self.type = AcquisitionType[acquisition.type.upper()]

                self.sensitivities = [Sensitivity(self)]
                for bench_name, value in self.parse_benches().items():
                    self.benches.append(
                        BenchFactory(
                            bench_name, value, self, model_file
                        )
                    )

    def parse_benches(self) -> dict[str, mBench]:
        """
        Parse metadata to get benches used in this acquisition
        """
        # WARNING: TO UPDATE WHEN ADDING BENCH
        if self.measure.metadata is None:
            raise Exception(_("associated measure's metadata not set"))
        results: dict[str, mBench] = dict()
        if self.measure.metadata.benches.SDB1 is not None:
            results["SDB1"] = self.measure.metadata.benches.SDB1
        if self.measure.metadata.benches.SDB2 is not None:
            results["SDB2"] = self.measure.metadata.benches.SDB2
        if self.measure.metadata.benches.SQB1 is not None:
            results["SQB1"] = self.measure.metadata.benches.SQB1
        if self.measure.metadata.benches.SNEB is not None:
            results["SNEB"] = self.measure.metadata.benches.SNEB
        if self.measure.metadata.benches.SWEB is not None:
            results["SWEB"] = self.measure.metadata.benches.SWEB
        if len(results) == 0:
            raise Exception(
                _("no benches associated to this acquisition")
            )
        return results

    def load_channel(self, channel: Channel) -> Signal:
        """
        Load data for a given channel

        Parameters
        ----------
        channel: Channel
            Channel to load
        """
        return channel.load(self.folder / (channel.name + ".csv"))

    def get_bench(self, bench_name: str) -> Bench:
        """
        Get a bench from this acquisition

        Parameters
        ----------
        bench_name: str
            Name of the bench to get
        """
        for bench in self.benches:
            if bench.name == bench_name:
                return bench
        raise Exception(
            _("bench {} does not exist in this measure").format(
                bench_name
            )
        )

    def show(self, axes: Axes, detailed: bool = False) -> Axes:
        """
        Show current acquisition on a given axe

        Parameters
        ----------
        axes: Axes
            Axes where to add graph
        detailed: bool
            If true, detail label of the plot
        """
        if self.sensitivities is None:
            raise Exception(
                _("cannot show sensitivity, as it is not defined")
            )
        if self.type is AcquisitionType.EXCITED:
            if self.projection is None:
                self.projection = Projection(
                    self,
                    {
                        "SNEB": 5.13e-9,
                        "SWEB": 2.62e-9,
                        "SDB1": 2.44e-9,
                        "SDB2": 1.62e-12,
                        "SQB1": 1.58e-12,
                    },
                )
            self.projection.compute()  # pyright: ignore[reportUnusedCallResult]
            axes = self.projection.show(axes, detailed)
        for sensitivity in self.sensitivities:
            axes = sensitivity.show(axes, detailed)
        return axes
