# pyright: reportImportCycles = false

from __future__ import annotations

from pathlib import Path

from science_signal import Signal  # pyright: ignore[reportMissingTypeStubs]
from .channel import ChannelType
from .coupling import Coupling

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .channel import Channel
    from .acquisition import Acquisition
    from .metadatas.bench import Bench as mBench


class Bench:
    """
    A bench in a measure

    Attributes
    ----------
    acquisition: Acquisition | None
        Acquisition that use this bench instance
    channels: set[Channel]
        Set of channels needed to retrieve this bench motion
    coupling: Coupling
        Coupling function associated to this bench for this acquisition
    metadata: metadatas.Bench | None
        Bench metadata for this acquisition
    movement: Signal | None
        Movement of this bench
    name: str
        Name of the bench
    """

    def __init__(
        self,
    ) -> None:
        """
        Initializer for Bench

        Parameters
        ----------
        measure: Measure
            Measure that use this bench instance
        """
        self.channels: set[Channel] = set()
        self.coupling: Coupling | None = None
        self.acquisition: Acquisition | None = None
        self.metadata: mBench | None = None
        self.movement: Signal | None = None
        self.name: str = ""

    def load(self) -> Signal:
        """
        Load channels data associated to this bench to compute movement
        of this bench. The data depends of the acquisition, so there is
        multiple movement (for multiple acquisition)
        """
        if self.acquisition is None:
            raise Exception(
                _(
                    "acquisition should be defined to load data from the bench"
                )
            )
        if self.metadata is None:
            raise Exception(
                _(
                    "calibration factors should be defined to load data from the bench"
                )
            )
        bench: None | Signal = None
        mirror: None | Signal = None
        for channel in self.channels:
            if channel.type is ChannelType.BENCH:
                if bench is not None:
                    raise Exception(
                        _("one bench can only accept one bench channel")
                    )
                bench = self.acquisition.load_channel(channel)
            elif channel.type is ChannelType.MIRROR:
                if mirror is not None:
                    raise Exception(
                        _(
                            "one bench can only accept one mirror channel"
                        )
                    )
                mirror = self.acquisition.load_channel(channel)
            else:
                raise Exception(
                    _(
                        "{} is an unknown channel type for a bench"
                    ).format(channel.type.name)
                )
        if bench is None:
            raise Exception(
                _("one bench must be associated to one bench channel")
            )
        if mirror is None:
            raise Exception(
                _("one bench must be associated to one mirror channel")
            )
        self.movement = (
            (
                bench * 1e-6 * self.metadata.calibration_bench
                - mirror * 1e-6 * self.metadata.calibration_mirror
            )
            .detrend("constant")
            .filter(end=1)
            .detrend("constant")
        )
        return self.movement

    def set_coupling(self, file: Path) -> Coupling:
        """
        Set coupling transfer function for this bench

        Parameters
        ----------
        file: Path
            Path to the coupling file where the transfer function are stored
        """
        self.coupling = Coupling(file, self)
        return self.coupling
