from __future__ import annotations

from enum import Enum
from pathlib import Path
from numpy import loadtxt
from science_signal import Signal, to_signal  # pyright: ignore[reportMissingTypeStubs]

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .bench import Bench


class ChannelType(Enum):
    """
    Types of channels
    """

    SENSITIVITY = 0
    BENCH = 1
    MIRROR = 2


class Channel:
    """
    A data channel from a specific acquisition

    Attributes
    ----------
    bench: Bench | None
        Bench associated to this channel. If no bench is associated to
        this channel (for h(t) signal for example), this value must be
        set to None
    data: Signal | None
        Data contained in this channel
    file: Path | None
        Path of this data channel in the filesystem
    name: str
        Name of the channel
    type: ChannelType
        Type of the channel
    """

    def __init__(
        self, name: str, type: ChannelType, bench: Bench | None = None
    ) -> None:
        """
        Initializer for Channel

        Parameters
        ----------
        name: str
            name of this channel
        type: ChannelType
            type of this channel
        bench: Bench | None
            Bench associated to this channel. If no bench is associated
            to this channel (for h(t) signal for example), this value
            must be set to None
        """
        if (type is ChannelType.SENSITIVITY) and (bench is not None):
            raise Exception(
                _(
                    "channel {} is a sensitivity channel and should not be associated to a bench"
                ).format(name)
            )
        self.name: str = name
        self.type: ChannelType = type
        self.bench: Bench | None = bench
        self.file: Path | None = None
        self.data: Signal | None = None

    def load(self, file: Path) -> Signal:
        """
        Set file attribute of this class and load the data from it

        Parameters
        ----------
        file: Path
            New file value
        """
        if not file.is_file():
            raise Exception(
                _("the file {} does not exist").format(file)
            )
        self.file = file
        self.data = self.load_data()
        return self.data

    def load_data(self) -> Signal:
        """
        Retrieve the data of this channel
        """
        if self.file is None:
            raise Exception(
                _(
                    "cannot load data from channel {} because the file associated to it is not defined"
                ).format(self.name)
            )
        return to_signal(loadtxt(self.file).T)
