# pyright: reportImportCycles=false
from __future__ import annotations

from pathlib import Path
from matplotlib.figure import Figure
from numpy import ceil

from .measure import Measure
from .metadata import Metadata


class Experiment:
    """
    Experiment with one (or no) injection of noise

    Attributes
    ----------
    folder: Path
        Folder of this experiment
    measures: list[Measure]
        Measure in this experiment
    metadata: Metadata
        Metadata of this experiment
    """

    def __init__(
        self,
        date: str,
        folder: Path,
        transfer_function: Path,
    ) -> None:
        """
        Initializer of Experiment class

        Parameters
        ----------
        date: str
            Date of the experiment as a string formatted YYYY_MM_DD.
            For example, "2024_10_31".
        folder: Path
            Data folder containing experiments
        transfer_function: Path
            mat file containing values of transfer function
        """
        self.folder: Path = (
            folder / "experiments" / "experiment_{}".format(date)
        )
        self.metadata: Metadata = Metadata(self.folder)
        self.measures: list[Measure] = []

        self.load(transfer_function)

    def load(self, transfer_function: Path) -> None:
        """
        Load experiment data from metadata

        Parameters
        ----------
        transfer_function: Path
            Path to a file containing transfer function that will be
            used to analyze this experiment
        """
        for measure in self.metadata.measures:
            self.measures.append(
                Measure(
                    self,
                    measure.name,
                    transfer_function,
                )
            )

    def get_measure(self, measure_name: str) -> Measure:
        """
        Get a bench from this experiment

        Parameters
        ----------
        bench_name: str
            Name of the bench to get
        """
        for measure in self.measures:
            if measure.name == measure_name:
                return measure
        raise Exception(
            _("measure {} does not exist in this experiment").format(
                measure_name
            )
        )

    def show(self, figure: Figure, detailed: bool = False) -> Figure:
        """
        Show current experiment on a given figure

        Parameters
        ----------
        figure: Figure
            Figure where to display this experiment
        detailed: bool
            If true, detail label of the plot
        """
        nb_cols: int = int(ceil(len(self.measures) ** (1 / 2)))  # pyright: ignore[reportAny]
        nb_rows: int = int(ceil(len(self.measures) / nb_cols))  # pyright: ignore[reportAny]
        for index in range(len(self.measures)):
            self.measures[index].show(
                figure.add_subplot(nb_rows, nb_cols, index + 1),
                detailed,
            )  # pyright: ignore[reportUnusedCallResult]
        return figure

    def fit_projections(self) -> dict[str, dict[str, float]]:
        """
        Fit projections for all measures in this experiment
        """
        results: dict[str, dict[str, float]] = dict()
        for measure in self.measures:
            results[measure.name] = measure.fit_projection()
        return results
