from __future__ import annotations

from typing import TYPE_CHECKING
from pathlib import Path

from scipy.constants import pi  # pyright: ignore[reportMissingTypeStubs]
from numpy import argmin, array, logspace, log10
from science_signal import Signal  # pyright: ignore[reportMissingTypeStubs]

from .utils import compute_light, get_factors
from .acquisition import Acquisition, AcquisitionType
from .metadatas.measure import Measure as mMeasure

if TYPE_CHECKING:
    from .experiment import Experiment
    from matplotlib.axes import Axes


class Measure:
    """
    A measure is a set of acquisitions and benches allowing to process
    multiple acquisitions at the same time (using common benches)

    Attributes
    ----------
    acquisitions: list[Acquisitions]
        Acquisition in this measure
    experiment: Experiment
        Experiment containing this measure
    folder: Path
        Folder containing this measure
    name: str
        Name of this measure
    metadata: metadatas.measure.Measure | None
        Metadata associated to this measure
    model_file: Path
        File containing the model used to compute projection
    """

    def __init__(
        self, experiment: Experiment, name: str, model_file: Path
    ) -> None:
        """
        Initializer of Measure

        Parameters
        ----------
        experiment: Experiment
            Experiment containing this measure
        """
        folder = experiment.folder / name

        if not folder.is_dir():
            raise Exception(_("{} is not a folder").format(folder))

        self.experiment: Experiment = experiment
        self.folder: Path = folder
        self.name: str = name

        self.metadata: mMeasure | None = None
        self.acquisitions: list[Acquisition] = []

        self.load(model_file)

    def get_acquisition(
        self,
        name: str | None = None,
        type: AcquisitionType | None = None,
    ) -> Acquisition:
        if name is not None:
            for acquisition in self.acquisitions:
                if acquisition.name == name:
                    if type is not None:
                        if acquisition.type != type:
                            raise Exception(
                                _(
                                    "the acquisition with the name {} is not a {} type"
                                ).format(name, type.name)
                            )
                    return acquisition
            raise Exception(
                _("acquisition with name {} not found").format(name)
            )
        elif type is not None:
            for acquisition in self.acquisitions:
                if acquisition.type == type:
                    return acquisition
            raise Exception(
                _("acquisition with type {} not found").format(
                    type.name
                )
            )
        raise Exception(
            _("cannot get an acquisition without a type or a name")
        )

    def load(self, model_file: Path) -> None:
        """
        Load attributes from metadata

        Parameters
        ----------
        model_file: Path
            File containing the model used to compute projection
        """
        for measure in self.experiment.metadata.measures:
            if measure.name == self.name:
                self.metadata = measure
                for acquisition in measure.acquisitions:
                    self.acquisitions.append(
                        Acquisition(self, acquisition.name, model_file)
                    )
                return None
        raise Exception(_("measure {} not found").format(self.name))

    def show(self, axes: Axes, detailed: bool = False) -> Axes:
        """
        Show current measure on a given subfigure

        Parameters
        ----------
        subfigure: SubFigure
            Subfigure where to display this measure
        detailed: bool
            If true, detail label of the plot
        """
        for acquisition in self.acquisitions:
            axes = acquisition.show(axes, detailed)
        axes.set_title("measure: {}".format(self.name))  # pyright: ignore[reportUnusedCallResult]
        axes.set_xlim(10, 100)  # pyright: ignore[reportUnusedCallResult]
        axes.set_ylim(10e-28, 10e-19)  # pyright: ignore[reportUnusedCallResult]
        axes.legend()  # pyright: ignore[reportUnusedCallResult]
        return axes

    def fit_projection(self) -> dict[str, float]:
        """
        Find the best backscatter factor for this measure

        This must be used with a measure which consist of excited and
        reference signal
        """
        if len(self.acquisitions) == 0:
            raise Exception(
                _(
                    "no acquisition in this measure, cannot fit projection"
                )
            )
        excited = self.acquisitions[0]
        reference = self.acquisitions[0]
        for acquisition in self.acquisitions:
            if acquisition.type is AcquisitionType.REFERENCE:
                reference = acquisition
            elif acquisition.type is AcquisitionType.EXCITED:
                excited = acquisition
            else:
                raise Exception(
                    _(
                        "no type for the acquisition {}".format(
                            acquisition.name
                        )
                    )
                )

        if excited.type is not AcquisitionType.EXCITED:
            raise Exception(
                _("no excited acquisition in the measure {}").format(
                    self.name
                )
            )
        if reference.type is not AcquisitionType.REFERENCE:
            raise Exception(
                _("no reference acquisition in the measure {}").format(
                    self.name
                )
            )

        if excited.sensitivities is None:
            raise Exception(
                _("no sensitivity defined in the excited acquisition")
            )
        if reference.sensitivities is None:
            raise Exception(
                _("no sensitivity defined in the reference acquisition")
            )
        if excited.sensitivities[0].data is None:
            raise Exception(
                _(
                    "no sensitivity data defined in the excited acquisition"
                )
            )
        if reference.sensitivities[0].data is None:
            raise Exception(
                _(
                    "no sensitivity data defined in the reference acquisition"
                )
            )

        phase = 4 * pi / 1.064e-6

        minimum_sum: dict[str, float] = dict()

        signal_excited = excited.sensitivities[0].get_psd()
        signal_reference = reference.sensitivities[0].get_psd()

        start = 15
        end = 100

        """
        First, we suppose the false assumption that every bench contribute
        alone to the bad sensitivity
        """
        for bench in excited.benches:
            """
            Fitting process
            """
            (
                factor_1,
                coupling_1,
                factor_2,
                coupling_2,
                signal_excited,
                signal_reference,
            ) = get_factors(
                bench,
                phase,
                excited.sensitivities[0].data.psd().sqrt(),
                reference.sensitivities[0].data.psd().sqrt(),
                start=start,
                end=end,
            )

            scatter_min = 1e-20
            scatter_max = 1e20

            min: float = 0

            for index in range(3):
                x = logspace(
                    log10(scatter_max),  # pyright: ignore[reportAny]
                    log10(scatter_min),  # pyright: ignore[reportAny]
                    1000,
                )
                y = array(
                    [
                        sum(
                            abs(
                                Signal(
                                    factor_1.x,
                                    compute_light(
                                        x[i] / 1e-6,
                                        factor_1,
                                        coupling_1,
                                        factor_2,
                                        coupling_2,
                                        phase,
                                    ),
                                )
                                + signal_reference
                                - signal_excited
                            ).y
                        )
                        for i in range(len(x))
                    ]
                )
                if argmin(y) == 0:
                    raise Exception(
                        _(
                            "best scatter factor seems below current range: {}".format(
                                scatter_min
                            )
                        )
                    )
                elif argmin(y) == len(x) - 1:
                    raise Exception(
                        _(
                            "best scatter factor seems above current range: {}".format(
                                scatter_max
                            )
                        )
                    )
                else:
                    scatter_min = x[argmin(y) - 1]
                    scatter_max = x[argmin(y) + 1]
                min = y[argmin(y)]

            if len(excited.benches) == 1:
                return {bench.name: (scatter_min + scatter_max) / 2}

            minimum_sum[bench.name] = min

        """
        We suppose the bench which can minimize the most the difference
        alone is the bench which contribute the most for the noise. Then, we
        can first take into account this bench, remove the associated
        computed projection to the global projection, and take into account
        the second most noisy bench, etc. until every bench is processed
        """

        scatter_factors: dict[str, float] = dict()

        for bench_name in sorted(  # pyright: ignore[reportCallIssue]
            minimum_sum,
            key=minimum_sum.get,  # pyright: ignore[reportArgumentType]
        ):  # lowest diff first
            """
            Same fitting process than before
            """
            (
                factor_1,
                coupling_1,
                factor_2,
                coupling_2,
                signal_excited,
                signal_reference,
            ) = get_factors(
                excited.get_bench(bench_name),
                phase,
                signal_excited,
                signal_reference,
                start=start,
                end=end,
            )

            scatter_min = 1e-20
            scatter_max = 1e0

            for index in range(3):
                x = logspace(
                    log10(scatter_max),  # pyright: ignore[reportAny]
                    log10(scatter_min),  # pyright: ignore[reportAny]
                    1000,
                )
                y = array(
                    [
                        sum(
                            abs(
                                Signal(
                                    factor_1.x,
                                    compute_light(
                                        x[i] / 1e-6,
                                        factor_1,
                                        coupling_1,
                                        factor_2,
                                        coupling_2,
                                        phase,
                                    ),
                                )
                                + signal_reference
                                - signal_excited
                            ).y
                        )
                        for i in range(len(x))
                    ]
                )
                if argmin(y) == 0:
                    raise Exception(
                        _(
                            "best scatter factor seems below current range: {}".format(
                                scatter_min
                            )
                        )
                    )
                elif argmin(y) == len(x) - 1:
                    raise Exception(
                        _(
                            "best scatter factor seems above current range: {}".format(
                                scatter_max
                            )
                        )
                    )
                else:
                    scatter_min = x[argmin(y) - 1]
                    scatter_max = x[argmin(y) + 1]

            scatter_factors[bench_name] = (
                scatter_min + scatter_max
            ) / 2
        return scatter_factors
