from pathlib import Path
from tomllib import load

from .metadatas.general import General
from .metadatas.benches import Benches
from .metadatas.acquisition import Acquisition
from .metadatas.bench import Bench
from .metadatas.measure import Measure


class Metadata:
    """
    Metadata accessor

    Attributes
    ----------
    general: General
        General section of metadata
    benches: Benches | None
        Benches section of metadata
    reference: Reference | None
        Reference section of metadata
    """

    def __init__(self, experiment_folder: Path) -> None:
        """
        Intializer of Metadata

        Parameters
        ----------
        experiment_folder: Path
            Folder containing data from various experiments (and the metadata file)
        """
        with open(experiment_folder / "metadata.toml.new", "rb") as f:
            data = load(f)

        self.general = General(**data["general"])
        self.measures: list[Measure] = []
        for measure in data["measures"]:
            # parse benches part
            dict_benches: dict[str, Bench] = dict()
            for bench_name, calibrations_value in measure[
                "benches"
            ].items():
                dict_benches[bench_name] = Bench(
                    calibrations_value["calibration_mirror"],
                    calibrations_value["calibration_mirror"],
                )
            benches = Benches(**dict_benches)

            # parse acquisitions part
            acquisitions: list[Acquisition] = []
            for acquisition_value in measure["acquisitions"]:
                acquisitions.append(
                    Acquisition(
                        acquisition_value["type"],
                        acquisition_value["name"],
                        acquisition_value["time"],
                        acquisition_value["duration"],
                        acquisition_value["frequency"],
                        acquisition_value["amplitude"],
                    )
                )

            self.measures.append(
                Measure(
                    measure["name"],
                    benches,
                    acquisitions,
                )
            )
