from __future__ import annotations

from typing import TYPE_CHECKING

from science_signal.signal import Signal  # pyright: ignore[reportMissingTypeStubs]
from scipy.constants import pi  # pyright: ignore[reportMissingTypeStubs]

from .utils import compute_light, get_factors

if TYPE_CHECKING:
    from .acquisition import Acquisition
    from matplotlib.axes import Axes


class Projection:
    """
    Projection for an acquisition.

    This class should not be used when managing batches of projections
    (e.g. when fitting projection). This would slow down the process
    considerably due to the overhead used here: each projection generate
    their own interpolated factor and couplings data.

    Attributes
    ----------
    data: Signal | None
        Value of the projection
    acquisition: Acquisition
        Acquisition associated to this projection
    scatter_factor: dict[str, float] | None
        Values of the scatter factor used for this projection for all
        benches, where keys are bench name
    """

    def __init__(
        self,
        acquisition: Acquisition,
        scatter_factor: dict[str, float] | None = None,
    ) -> None:
        """
        Initalizer of Projection

        Parameters
        ----------
        experiment: Experiment
            Experiment associated to this projection
        scatter_factor: dict[str, float] | None
            Scatter factor of this projection
        """
        if scatter_factor is None:
            scatter_factor = {
                "SDB1": 1e-9,
                "SDB2": 1e-9,
                "SNEB": 1e-9,
                "SWEB": 1e-9,
                "SQB1": 1e-9,
            }
        self.acquisition: Acquisition = acquisition
        self.scatter_factor: dict[str, float] = scatter_factor
        self.data = None
        self.signals: list[Signal] = []

    def compute(self) -> Signal:
        """
        Compute the projection for a given scatter_factor
        """
        if len(self.acquisition.benches) == 0:
            raise Exception(
                _(
                    "no bench in this experiment, cannot compute projection"
                )
            )
        signals: list[Signal] = []
        for bench in self.acquisition.benches:
            phase = 4 * pi / 1.064e-6
            (
                factor_1,
                coupling_1,
                factor_2,
                coupling_2,
            ) = get_factors(bench, phase)
            signals.append(
                Signal(
                    factor_1.x,
                    compute_light(
                        self.scatter_factor[bench.name]
                        / 1e-6,  # divide by the reflectivity of the mirror (equivalent bench) in the model
                        factor_1,
                        coupling_1,
                        factor_2,
                        coupling_2,
                        phase,
                    ),
                )
            )
        result = signals[0]
        self.signals = signals
        for signal in signals[1:]:
            result += signal
        self.data = result
        return self.data

    def show(self, axes: Axes, detailed: bool = False) -> Axes:
        """
        Show current projection on a given axes

        Parameters
        ----------
        axes: Axes
            Axes where to add graph
        detailed: bool
            If true, detail label of the plot
        """
        if self.data is None:
            raise Exception(
                _(
                    "cannot show projection because its value is not computed"
                )
            )
        if detailed:
            label = _("{}: {}-{} {} projection").format(
                self.acquisition.measure.experiment.metadata.general.name,
                self.acquisition.measure.name,
                self.acquisition.name,
                "{}",
            )
        else:
            label = _("{} projection")
        for i in range(len(self.signals)):
            signal = self.signals[i]
            axes.loglog(
                signal.x,
                signal.y,
                label=_(label).format(self.acquisition.benches[i].name),
            )  # pyright: ignore[reportUnusedCallResult]
        return axes
