from __future__ import annotations

from numpy import loadtxt
from pathlib import Path

from .channel import Channel, ChannelType
from science_signal import Signal  # pyright: ignore[reportMissingTypeStubs]

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from .acquisition import Acquisition


class Sensitivity:
    """
    A detected sensitivity for an experiment

    Attributes
    ----------
    channel: Channel
        Channel associated to this sensitivity
    data: Signal | None
        Sensitivity data
    acquisition: Acquisition
        Acquisition associated to this sensitivity
    already_processed: bool
        If sensitivity is already in frequency domain, set to True
    """

    def __init__(
        self, acquisition: Acquisition, from_txt: Path | None = None
    ) -> None:
        """
        Initializer for Sensitivity

        Parameters
        ----------
        experiment: Experiment
            Experiment that use this sensitivity instance
        from_txt: Path | None
            Path of a text file to load this sensitivity values
        """
        self.acquisition: Acquisition = acquisition
        self.channel: Channel = Channel("h", ChannelType.SENSITIVITY)
        self.data: Signal | None = None
        self.already_processed: bool = False

        if from_txt is None:
            self.load()  # pyright: ignore[reportUnusedCallResult]
        else:
            data = loadtxt(from_txt)
            self.data = Signal(data[:, 0], data[:, 1])
            self.already_processed = True

    def load(self) -> Signal:
        """
        Load channel data associated to this sensitivity
        """
        self.data = self.acquisition.load_channel(self.channel)
        return self.data

    def get_psd(self) -> Signal:
        """
        Return the square of the psd of this sensitivity
        """
        if self.data is None:
            raise Exception(_("data should be defined to process it"))
        return self.data.psd().sqrt()

    def show(self, axes: Axes, detailed: bool = False) -> Axes:
        """
        Show current sensitivity on a given axes

        Parameters
        ----------
        axes: Axes
            Axes where to add graph
        detailed: bool
            If true, detail label of the plot
        """
        if self.already_processed:
            if self.data is None:
                raise Exception(_("data should be defined to be shown"))
            signal = self.data
        else:
            signal = self.get_psd()
        if detailed:
            label = _("{}: {}-{} sensitivity").format(
                self.acquisition.measure.experiment.metadata.general.name,
                self.acquisition.measure.name,
                self.acquisition.name,
            )
        else:
            label = _("sensitivity")
        axes.loglog(
            signal.x,
            signal.y,
            label=label,
        )  # pyright: ignore[reportUnusedCallResult]
        return axes
