# pyright: reportAny = false
from __future__ import annotations

from math import log10
from science_signal.signal import Signal, interpolate  # pyright: ignore[reportMissingTypeStubs]
from typing import TYPE_CHECKING
from numpy.typing import NDArray
from numpy import argmin, array, float64, logspace, sqrt
from scipy.constants import pi  # pyright: ignore[reportMissingTypeStubs]
from .bench import Bench

if TYPE_CHECKING:
    from .acquisition import Acquisition


def compute_light(
    pre_scatter_factor: float,
    factor_1: Signal,
    coupling_1: Signal,
    factor_2: Signal,
    coupling_2: Signal,
    phase: float,
) -> NDArray[float64]:
    """
    Optimized computing of light with pre-computed factor

    Parameters
    ----------
    pre_scatter_factor: float
        Scatter factore value CORRECTED (taking account of the
        reflection of the mirror (equivalent to bench) used in the model
    factor_1: Signal
        Movement factor, which contains the Fourier transform of the
        sine of the bench's movement
    coupling_1: Signal
        Transfer function values, for the phase part
    factor_2: Signal
        Movement factor, which contains the Fourier transform of the cos
        of the bench's movement
    coupling_2: Signal
        Transfer function values, for the radiation pressure part
    phase: float
        Phase part, which is used despite only being in the sine and cos
        in the formula 6.5 on the page 162 of the thesis
        [https://theses.hal.science/tel-04124206v1](here). The phase
        is needed again because ? I'M NOT SURE !!!!
    """
    return (
        sqrt(pre_scatter_factor)
        / phase
        * (coupling_1 * factor_1 + coupling_2 * factor_2).y
    )


def get_factors(
    bench: Bench,
    phase: float,
    *signals: Signal,
    start: float | None = None,
    end: float | None = None,
) -> tuple[Signal, ...]:
    """
    Get interpolated factors and coupling to compute a projection for a
    given scatter factor

    Parameters
    ----------
    bench: Bench
        Bench to get factors from (use coupling and movement from this
        bench)
    phase: float
        Phase of the light allowing to encode the relative motion between
        the mirror and the bench
    *signals: Signal
        Signal which will be interpolated with the returned factor for
        convenience (only interpolate theses)
    start: float | None
        Start frequency to interpolate with (allows to remove low
        frequence)
    end: float | None
        End frequency to interpolate with (allows to remove high
        frequence)
    """
    if bench.coupling is None:
        raise Exception(
            _(
                "cannot get factors when the bench does not have associated transfer function"
            )
        )
    if bench.movement is None:
        raise Exception(
            _(
                "cannot get factors when the bench movement is not computed"
            )
        )

    coupling_1 = bench.coupling.data[0]
    coupling_2 = bench.coupling.data[1]

    factor_1 = (bench.movement * phase).sin().psd().sqrt()
    factor_2 = (bench.movement * phase).cos().psd().sqrt()

    if start is None:
        start = coupling_1.x[0]
    if end is None:
        end = coupling_1.x[-1]

    coupling_1 = coupling_1.cut(start, end)

    return interpolate(
        factor_1,
        coupling_1,
        factor_2,
        coupling_2,
        *signals,
    )


def fit_projection(
    excited: Acquisition, reference: Acquisition
) -> tuple[dict[str, float], Signal]:
    """
    Find the best backscatter factor

    Find the best scatter factor (corrected with reflection of the
    mirror in the model) which minimize the difference between the sum
    of the projection and the reference sensitivity and the excited
    sensitivity

    Parameters
    ----------
    excited: Acquisition
        Excited acquisition
    reference: Acquisition
        Reference acquisition
    """
    from .acquisition import AcquisitionType

    if excited.type is not AcquisitionType.EXCITED:
        raise Exception(
            _(
                "excited acquisition should be an excited acquisition, where an injection was done"
            )
        )
    if reference.type is not AcquisitionType.REFERENCE:
        raise Exception(
            _(
                "reference acquisition should be a reference acquisition, without any injection"
            )
        )
    if excited.sensitivity is None:
        raise Exception(
            _("no sensitivity defined in the excited acquisition")
        )
    if reference.sensitivity is None:
        raise Exception(
            _("no sensitivity defined in the reference acquisition")
        )
    if excited.sensitivity.data is None:
        raise Exception(
            _("no sensitivity data for excited acquisition")
        )
    if reference.sensitivity.data is None:
        raise Exception(
            _("no sensitivity data for reference acquisition")
        )

    phase = 4 * pi / 1.064e-6

    minimum_sum: dict[str, float] = dict()

    signal_excited = excited.sensitivity.data.psd().sqrt()
    signal_reference = reference.sensitivity.data.psd().sqrt()

    start = 15
    end = 100

    """
    First, we suppose the false assumption that every bench contribute
    alone to the bad sensitivity
    """
    for bench in excited.benches:
        """
        Fitting process
        """
        (
            factor_1,
            coupling_1,
            factor_2,
            coupling_2,
            signal_excited,
            signal_reference,
        ) = get_factors(
            bench,
            phase,
            excited.sensitivity.data.psd().sqrt(),
            reference.sensitivity.data.psd().sqrt(),
            start=start,
            end=end,
        )

        scatter_min = 1e-20
        scatter_max = 1e20

        min: float = 0

        for index in range(3):
            x = logspace(log10(scatter_max), log10(scatter_min), 1000)
            y = array(
                [
                    sum(
                        abs(
                            Signal(
                                factor_1.x,
                                compute_light(
                                    x[i] / 1e-6,
                                    factor_1,
                                    coupling_1,
                                    factor_2,
                                    coupling_2,
                                    phase,
                                ),
                            )
                            + signal_reference
                            - signal_excited
                        ).y
                    )
                    for i in range(len(x))
                ]
            )
            if argmin(y) == 0:
                raise Exception(
                    _(
                        "best scatter factor seems below current range: {}".format(
                            scatter_min
                        )
                    )
                )
            elif argmin(y) == len(x) - 1:
                raise Exception(
                    _(
                        "best scatter factor seems above current range: {}".format(
                            scatter_max
                        )
                    )
                )
            else:
                scatter_min = x[argmin(y) - 1]
                scatter_max = x[argmin(y) + 1]
            min = y[argmin(y)]

        if len(excited.benches) == 1:
            return {
                bench.name: (scatter_min + scatter_max) / 2
            }, signal_excited - Signal(
                factor_1.x,
                compute_light(
                    (scatter_min + scatter_max) / 2,
                    factor_1,
                    coupling_1,
                    factor_2,
                    coupling_2,
                    phase,
                ),
            )
        minimum_sum[bench.name] = min

    """
    We suppose the bench which can minimize the most the difference
    alone is the bench which contribute the most for the noise. Then, we
    can first take into account this bench, remove the associated
    computed projection to the global projection, and take into account
    the second most noisy bench, etc. until every bench is processed
    """

    scatter_factors: dict[str, float] = dict()

    for bench_name in sorted(  # pyright: ignore[reportCallIssue]
        minimum_sum,
        key=minimum_sum.get,  # pyright: ignore[reportArgumentType]
    ):  # lowest diff first
        """
        Same fitting process than before
        """
        (
            factor_1,
            coupling_1,
            factor_2,
            coupling_2,
            signal_excited,
            signal_reference,
        ) = get_factors(
            excited.get_bench(bench_name),
            phase,
            signal_excited,
            signal_reference,
            start=start,
            end=end,
        )

        scatter_min = 1e-20
        scatter_max = 1e0

        for index in range(3):
            x = logspace(log10(scatter_max), log10(scatter_min), 1000)
            y = array(
                [
                    sum(
                        abs(
                            Signal(
                                factor_1.x,
                                compute_light(
                                    x[i] / 1e-6,
                                    factor_1,
                                    coupling_1,
                                    factor_2,
                                    coupling_2,
                                    phase,
                                ),
                            )
                            + signal_reference
                            - signal_excited
                        ).y
                    )
                    for i in range(len(x))
                ]
            )
            if argmin(y) == 0:
                raise Exception(
                    _(
                        "best scatter factor seems below current range: {}".format(
                            scatter_min
                        )
                    )
                )
            elif argmin(y) == len(x) - 1:
                raise Exception(
                    _(
                        "best scatter factor seems above current range: {}".format(
                            scatter_max
                        )
                    )
                )
            else:
                scatter_min = x[argmin(y) - 1]
                scatter_max = x[argmin(y) + 1]

        signal_excited -= compute_light(  # remove the projection from the excited sensitivity
            (scatter_min + scatter_max) / 2 / 1e-6,
            factor_1,
            coupling_1,
            factor_2,
            coupling_2,
            phase,
        )
        scatter_factors[bench_name] = (scatter_min + scatter_max) / 2
    return scatter_factors, Signal(factor_1.x, signal_excited)  # pyright: ignore[reportArgumentType, reportPossiblyUnboundVariable]
